import datetime
import asyncio
import pytest


def test_schedule_of_scheduler_method(scheduler, async_function):
    # Decorator
    task = scheduler.schedule(interval=datetime.timedelta(minutes=5))(async_function)

    assert task in scheduler.tasks
    assert task.task == async_function
    assert task.interval == datetime.timedelta(minutes=5)


def test_exception_in_schedule_method(scheduler):

    def test_function():
        print("hello world")

    with pytest.raises(ValueError) as err:
        # Decorator over the isn't couroutine function
        task = scheduler.schedule(interval=datetime.timedelta(minutes=5))(test_function)
    assert "Function under decorator must be await. Use 'async def' syntax" in str(err.value)


def test_run_in_scheduler(scheduler):
    # run without tasks
    assert scheduler.run() == None

def test_run_without_repeat(scheduler, scheduled_task):
    scheduled_task.repeat = False
    assert scheduler.run() == None


def test_repr_in_task(scheduled_task, async_function):
    assert repr(scheduled_task) in f"<SchedulerTask task={async_function.__name__} every {scheduled_task.interval}"


def test_immediate_attr_in_task(scheduler, async_function):
    task = scheduler.schedule(interval=datetime.timedelta(minutes=5), immediate=True)(async_function)
    # Test that if arg immediate True task have to start immediatly
    difference_of_time = (task.time_to - datetime.datetime.now()).total_seconds()
    assert int(difference_of_time) == 0


def test_time_to_is_time_now_plus_interval_in_task(scheduled_task):
    time_start = datetime.datetime.now()
    setted_time = scheduled_task.set_time()
    difference_of_time = (setted_time-time_start).total_seconds()
    # Test that set_time method set time to like datetime.datetime.now() + interval
    assert int(difference_of_time) == int(scheduled_task.interval.total_seconds())
    assert scheduled_task.time_to == setted_time


@pytest.mark.asyncio
async def test_time_to_do_in_task(scheduled_task):
    scheduled_task.interval = datetime.timedelta(seconds=1)
    scheduled_task.set_time()
    # test that time_to_do return right bool answers
    assert not scheduled_task.time_to_do()
    await asyncio.sleep(1)
    assert scheduled_task.time_to_do()


@pytest.mark.asyncio
async def test_wait_for_interval_method_in_task(scheduled_task):
    scheduled_task.interval = datetime.timedelta(seconds=1)
    # Test that wait_fo_interval raise error
    with pytest.raises(RuntimeError) as err:
        await scheduled_task.wait_for_interval()
    assert "set_time() must be called" in str(err.value)

    time_start = datetime.datetime.now()
    scheduled_task.set_time()
    # Test that wait_fo_interval wait correctly
    await scheduled_task.wait_for_interval()
    difference_of_time = (datetime.datetime.now()-time_start).total_seconds()
    assert int(difference_of_time) == int(scheduled_task.interval.total_seconds())


@pytest.mark.asyncio
async def test_iter_count_without_repeat(scheduler, async_function):
    with pytest.raises(ValueError) as err:
        scheduler.schedule(interval=datetime.timedelta(seconds=1), repeat=False, iter_count=5)(async_function)

    assert "If you want to repeat set arg repeat=True" in str(err.value)
    task = scheduler.schedule(interval=datetime.timedelta(seconds=1), iter_count=5)(async_function)
    assert task.repeat == True
    assert task.iter_count == 5


def test_next_iter(scheduled_task):
    task_counts = scheduled_task.count
    scheduled_task.next_iter()
    assert scheduled_task.count == task_counts + 1


def test_simple_task_with_iter_count(scheduler, async_function, capsys):

    task = scheduler.schedule(interval=datetime.timedelta(seconds=1), iter_count=2)(async_function)

    assert scheduler.run() == None
    assert "Hello world" in capsys.readouterr().out.strip()


def test_create_task_witout_decorator(scheduler, async_function, capsys):
    task = scheduler.register_task(async_function)
    with pytest.raises(ValueError) as err:
        scheduler.run()

    assert "Task must be scheduled before running. Use task.schedule(interval: datetime.timedelta)" in str(err.value)

    task.schedule(interval=datetime.timedelta(seconds=5), repeat=False)
    assert scheduler.run() == None
    assert "Hello world" in capsys.readouterr().out.strip()

def test_create_task_with_args_without_decorator(scheduler, capsys):

    async def test_args(a: int, *args, b: str = "some", **kwargs):
        print("a")
        print(args)
        for i in args:
            print(f"arg in args: {i}")
        print(b)
        print(kwargs)
        for i in kwargs:
            print(f"{i} = {kwargs[i]}")

    task = scheduler.register_task(test_args, 5, 3, 4, 23, 3, b="yes", name="Robert", age=25)

    with pytest.raises(ValueError) as err:
        scheduler.run()

    assert "Task must be scheduled before running. Use task.schedule(interval: datetime.timedelta)" in str(err.value)
    task.schedule(interval=datetime.timedelta(seconds=5), repeat=False)
    assert scheduler.run() == None
    out = capsys.readouterr().out.strip()
    assert "age = 25" in out
    assert "name = Robert" in out
    assert "5" in out
    assert "arg in args: 23" in out
    assert "yes" in out


def test_running_task_in_existing_event_loop(scheduler, scheduled_task, capsys):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    scheduled_task.repeat = False
    assert scheduler.run() == None
    assert "Hello world" in capsys.readouterr().out.strip()



import sys
import click
from importlib.util import spec_from_file_location, module_from_spec


sys.path.append(".")

class CustomPath(click.Path):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx) -> tuple:
        """
            Validate and check if path exist
        """
        splited_path = value.split(":")
        path, obj = splited_path if len(splited_path) == 2 else [None, None]
        if not path or not obj:
            raise ValueError("Incorrect Type of path to object")
        filename = path.split("/")[-1]
        converted_path = super().convert(path + ".py", param, ctx)
        
        return converted_path, filename, obj


@click.command()
@click.argument("app", type=CustomPath(exists=True))
def skd(app):
 
    """Start work task's scheduler

        \b
        Example:
            my_project/src/run:app
        \b
        Where:
            my_project/src - path to module,
            run.py - your module,
            app - object of Scheduler class"
            
    """
    path, name, obj = app
    

    # Create a spec for the module from the given file path
    spec = spec_from_file_location(name, path)
    # Create the module from the spec
    module = module_from_spec(spec)
    # Load and execute the module
    spec.loader.exec_module(module)

    # Start Scheduler
    scheduler = getattr(module, obj)
    scheduler.run()
    click.echo("Tasks completed successfully")

if __name__ == "__main__": # pragma: no cover
    skd()


from aioskd.cli import skd


def test_incorect_type_of_path_skd(click_CLI):
    result = click_CLI.invoke(skd, ["some.path.to.obj"])
    assert isinstance(result.exception, ValueError)
    assert "Incorrect Type of path to object" in str(result.exception)
    result = click_CLI.invoke(skd, ["some/path/to/obj"])
    assert isinstance(result.exception, ValueError)
    assert "Incorrect Type of path to object" in str(result.exception)


def test_not_exist_path(click_CLI):
    result = click_CLI.invoke(skd, ["some/path/to/obj:test"])

    assert isinstance(result.exception, SystemExit)
    assert "Error: Invalid value for 'APP': Path 'some/path/to/obj.py' does not exist." in str(result.output)


def test_exsist_path_but_incorent_path_to_obj(click_CLI, tmp_path):
    path = tmp_path / "test.py"
    path = f"{str(path).replace('.py', '')}"
    result = click_CLI.invoke(skd, [path])

    assert isinstance(result.exception, ValueError)
    assert "Incorrect Type of path to object" in str(result.exception)


def test_exsist_path(click_CLI, tmp_path):
    path = tmp_path / "test.py"
    path.write_text("""from aioskd.scheduler import Scheduler\nskd = Scheduler()""")

    path = f"{str(path).replace('.py', '')}"
    result = click_CLI.invoke(skd, [ f"{path}:skd" ])
    assert "Tasks completed successfully" in str(result.output)


def test_file_with_task(click_CLI, tmp_path):
    path = tmp_path / "test.py"
    path.write_text("""import datetime\nfrom aioskd.scheduler import Scheduler\nskd = Scheduler()\n@skd.schedule(interval=datetime.timedelta(seconds=1), repeat=False)\nasync def test_func():\n\tprint("hello world")""")
    path = f"{str(path).replace('.py', '')}"
    result = click_CLI.invoke(skd, [ f"{path}:skd" ])
    output = str(result.output)
    assert "Tasks completed successfully" in str(output)
    assert "hello world" in str(output)



from setuptools import setup, find_packages


def get_readme():
    with open("README.md", "r") as file:
        return file.read()


setup(
  name='aioskd',
  version='0.0.8',
  author='Artem Sydorenko',
  author_email='kradworkmail@gmail.com',
  description='Tool for execute async background tasks',
  long_description=get_readme(),
  long_description_content_type='text/markdown',
  packages=find_packages(),
  install_requires=['click==8.1.3'],
  classifiers=[
    'Programming Language :: Python :: 3.11'
  ],
  entry_points='''
        [console_scripts]
        skd=aioskd.cli:skd
  ''',
  keywords='async, scheduling, background-tasks, asynchronous-programming, scheduler-library, python, task-scheduler, background-processing, concurrency, asyncio, timed-tasks, library, python3, interval-tasks, task-scheduling, asynchronous-background-tasks, background-scheduler, background-jobs, background-execution, task-execution, scheduling-tasks, async-jobs, async-tasks, job-scheduler, task-runner, task-manager',
  python_requires='>=3.11'
)


import asyncio
import datetime
from typing import Callable
import functools
from inspect import iscoroutinefunction


class SchedulerTask:
    """
        A class that displays the task of the scheduler
        :param task: Asynchronous function to be executed
        :param interval: The interval between which the asynchronous function should be executed
        :param repeat: Indicates whether the asynchronous function should be repeated or executed once
        :param immediate: If yes, the scheduler will not wait for the interval when starting the task for the FIRST TIME
        :param iter_count: necessary number of repeats

    """
    def __init__(
            self,
            task: Callable,
            interval: datetime.timedelta | None = None,
            immediate: bool = False,
            repeat: bool = True,
            iter_count: int | None = None) -> None:
        
        self.task = task
        self.interval = interval
        self.count = 0
        self.repeat = repeat
        self.iter_count = iter_count
        if self.iter_count and not self.repeat:
            raise ValueError("If you want to repeat set arg repeat=True")

        self.immediate = immediate
        if self.immediate:
            self.time_to = datetime.datetime.now()
        else:
            self.time_to = None

    def schedule(
            self,
            interval: datetime.timedelta,
            immediate: bool = False,
            repeat: bool = True,
            iter_count: int | None = None) -> None:
        
        self.__init__(task=self.task, interval=interval, immediate=immediate, repeat=repeat, iter_count=iter_count)

    def next_iter(self):
        self.count += 1
        self.set_time()

    def set_time(self) -> datetime.datetime:
        """
            Sets the time at which the self.task should be executed
        """
        if not self.interval:
            raise ValueError("Task must be scheduled before running. Use task.schedule(interval: datetime.timedelta)")

        self.time_to = datetime.datetime.now() + self.interval
        return self.time_to

    async def wait_for_interval(self) -> None:
        if not self.time_to:
            raise RuntimeError("set_time() must be called")
        
        amount_of_time = self.time_to - datetime.datetime.now()
        await asyncio.sleep(amount_of_time.total_seconds())

    def time_to_do(self) -> bool:
        """
            Returns True if it is time to execute self.task
        """
        if self.time_to and datetime.datetime.now() >= self.time_to:
            return True
        return False

    def __repr__(self):
        return f"<SchedulerTask task={self.task.__name__} every {self.interval}"


class Scheduler:
    """
        A scheduler class that is used to perform background tasks in another process.
        
        Problem: 
            I had an asynchronous parsing system that I was thinking of using with celery beat, but celery doesn't work with asyncio yet.
            Unfortunately, I did not find a ready-made solution that would suit me on the Internet, so I decided to implement a scheduler for my own requirements
        
        Solution:
            Create a tool that allows you to run background tasks
            Requirements:
                Support for asyncio 
                Support for concurrent
                Independence from external libraries
                Convenient interaction interface with decorators
    """
    def __init__(self):
        self.tasks = []

    async def _execute_task(self, task: SchedulerTask) -> None:
        """
            A coroutine that checks if it is time to execute the task, if so, executes it, if the task needs to be executed once (repeat), interrupts it, if not, sets a new time
            :param task: The task to be executed
        """
        while True:
            if task.time_to_do():
                await task.task()
                if not task.repeat or task.count == task.iter_count:
                    break
                task.next_iter()

            await task.wait_for_interval()

    def _init_tasks(self) -> None:
        """
           A method that sets the execution time for all tasks that require it
        """
        [task.set_time() for task in self.tasks if not task.time_to]

    async def _run_with_interval(self) -> None:
        """
            Coroutine that creates tasks in which the scheduler tasks will be performed
        """
        self._init_tasks()
        tasks = [asyncio.create_task(self._execute_task(task)) for task in self.tasks]
        await asyncio.gather(*tasks)

    def schedule(
            self,
            interval: datetime.timedelta,
            immediate: bool = False,
            repeat: bool = True, iter_count: int | None = None) -> Callable:
        """
            Decorator that collects all the coroutines that need to be executed with time scheduling

            :param interval: The interval between which the asynchronous function should be executed
            :param repeat: Indicates whether the asynchronous function should be repeated or executed once
            :param immediate: If yes, the scheduler will not wait for the interval when starting the task for the FIRST TIME
            :param iter_count: necessary number of repeats
        """
        def wrapper(func) -> None:
            """
                :param func: Asynchronous function to be performed
            """
            if not iscoroutinefunction(func):
                raise ValueError("Function under decorator must be await. Use 'async def' syntax")
            task = SchedulerTask(task=func, interval=interval, repeat=repeat, immediate=immediate, iter_count=iter_count)
            self.tasks.append(task)
            return task

        return wrapper
    
    def register_task(self, task: Callable, *args, **kwargs) -> SchedulerTask:
        """
            Register a new task without scheduling it.

            This method allows you to manually create a task without using the 'schedule' decorator.
            The task can later be scheduled

            :param task: The asynchronous function to be performed as the task.
            :param *args: Positional arguments to be passed to the task function when it is executed.
            :param **kwargs: Keyword arguments to be passed to the task function when it is executed.
            :return: A SchedulerTask object representing the registered task.
            :rtype: SchedulerTask
        """
        partial_task = functools.partial(task, *args, **kwargs)
        functools.update_wrapper(partial_task, task)
        
        new_task = SchedulerTask(task=partial_task)
        self.tasks.append(new_task)
        return new_task

    def run(self) -> None:
        try:
            loop = asyncio.get_running_loop()
            loop.create_task(self._run_with_interval())
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self._run_with_interval())
